/*
* Copyright (c) 2007, Shenyang Institute of Automation, Chinese Academy
* of Sciences. , all rights reserved.
* 
* Filename     : MeshSegmentation.cpp
* Project	   : xMesh library
* Abstract     : Implementation of the MeshSegmentation class class.
* Version      : 1.0
* Author			: , nanlliang@sia.cn
* Original Author   : , nanlliang@sia.cn
* Accomplished date : Dec. 29, 2007
*/

//-----------------------------------------------------------------
#include "stdafx.h"
#include "MeshSegmentation.h"
#include "Constant.h"

#include <cassert>
#include <algorithm>
#include <stack>


MeshSegmentation::MeshSegmentation()
: m_pMesh(NULL)
, m_bShowSegmentation(true)
, m_listSegmentation(0)
, m_nNumComponents(0)
{}

MeshSegmentation::~MeshSegmentation(void)
{}

void MeshSegmentation::SetTarget(TriMesh* pMesh)
{
	m_pMesh = pMesh;
}

void MeshSegmentation::Clear()
{
	m_pMesh = NULL;
	m_faceID.clear();
	m_nNumComponents = 0;

	::glDeleteLists(m_listSegmentation, 1);
}


void MeshSegmentation::SetShowSegmentation(bool b)
{
	m_bShowSegmentation = b;
}

void MeshSegmentation::Apply(FeatureCriterion criterion, float value)
{
	DetectFeatureEdges(criterion, value);

	std::vector<TriFace>&   facets = m_pMesh->m_faces;

	m_faceID.clear();
	m_faceID.resize(facets.size(), -1);
	EnumerateComponents(m_faceID);

	BuildSegmentationList();
}

void MeshSegmentation::Draw()
{
	if ((!m_bShowSegmentation) || (!m_pMesh) || (m_faceID.empty()))
		return;

	// Χ
	Point3D center = m_pMesh->m_center;
	double fitScale = m_pMesh->m_fitScale;
	
	::glPushMatrix();
	::glScaled(fitScale, fitScale, fitScale);
	::glTranslated(-center.x, -center.y, -center.z);	

	::glCallList(m_listSegmentation);
	::glPopMatrix();
}


void MeshSegmentation::BuildSegmentationList()
{
	std::vector< std::vector<int> > allComponents(m_nNumComponents);
	for (int i=0; i<(int)m_faceID.size(); ++i)
		allComponents[ m_faceID[i] ].push_back(i);

	int i0 = 0; int i1 = 0; int i2 = 0;

	int faceNum = m_pMesh->m_faceNum;
	std::vector<TriFace>& faces = m_pMesh->m_faces;
	std::vector<Point3D>& vertices = m_pMesh->m_vertices;

	m_listSegmentation = glGenLists(1);
	::glNewList(m_listSegmentation, GL_COMPILE);
		::glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_LIGHTING_BIT);
		::glDisable(GL_LIGHTING);
		::glShadeModel(GL_FLAT);
		::glBegin(GL_TRIANGLES);
		for(int i=0; i<m_nNumComponents; i++) {
			std::vector<int>& oneComponent = allComponents[i];	
			GLuint r = rand()%256;
			GLuint g = rand()%256;
			GLuint b = rand()%256;
			::glColor3ub(r, g, b);
			for (int j=0; j<(int)oneComponent.size(); ++j)
			{
				int faceIndex = oneComponent[j];
				i0 = faces[faceIndex].index0;	i1 = faces[faceIndex].index1;	i2 = faces[faceIndex].index2;
				if (i0 < 0 || i1 < 0 || i2 < 0)
					continue;
				::glVertex3d(vertices[i0].x, vertices[i0].y, vertices[i0].z);
				::glVertex3d(vertices[i1].x, vertices[i1].y, vertices[i1].z);
				::glVertex3d(vertices[i2].x, vertices[i2].y, vertices[i2].z);
			}
		}
		::glEnd();
		::glPopAttrib();
	::glEndList();
}

void MeshSegmentation::DetectFeatureEdges(FeatureCriterion criterion, float value)
{
	m_pMesh->m_featureEdges.clear();

	if (m_pMesh->m_edges.empty())
		m_pMesh->ExtractEdges();
	if (m_pMesh->m_faceNormals.empty())
		m_pMesh->ComputeFaceNormal();

	std::set<Edge3D>& edges = m_pMesh->m_edges;
	std::set<Edge3D>::iterator cur = edges.begin();
	std::set<Edge3D>::iterator end = edges.end();

	if (criterion == FC_RATIO)
	{
		typedef std::pair<double, Edge3D>  aePair;
		std::vector<aePair> angles_and_edges;

		for (; cur!=end; ++cur)
		{
			if (!cur->isBorder)
				angles_and_edges.push_back(aePair( FaceNormalProduct(*cur), *cur ));
		}

		std::sort(angles_and_edges.begin(), angles_and_edges.end());

		// ȡ5%
		int size  = static_cast<int>(angles_and_edges.size() * value);
		m_pMesh->m_featureEdges.resize(size);
		for (int i=0; i<size; i++)
			m_pMesh->m_featureEdges[i] = angles_and_edges[i].second;
	}
	else if (criterion == FC_DIHEDRAL)
	{
		for (; cur!=end; ++cur)
		{
			if ( (!cur->isBorder) && (FaceNormalAngle(*cur) >= value) )
				m_pMesh->m_featureEdges.push_back(*cur);
		}
	}
	else
		AfxMessageBox(_T("Ч׼!"));
}

void MeshSegmentation::EnumerateComponents(std::vector<int>&  faceID) 
{
	std::set<Edge3D>        featureEdges;
	featureEdges.insert(m_pMesh->m_featureEdges.begin(), m_pMesh->m_featureEdges.end());

	std::vector<TriFace>&   facets = m_pMesh->m_faces;

	int num_facets = (int)facets.size();

	std::map<TriFace, int>   faceIndices;
	for (int i=0; i<num_facets; ++i)
		faceIndices[ facets[i] ] = i;

	m_nNumComponents = 0;
	for (int i=0; i<num_facets; ++i)
	{
		if (faceID[i] == -1) {
			PropagateComponent(facets[i], faceIndices, featureEdges, m_nNumComponents) ;
			++m_nNumComponents;
		}
	}
}


void MeshSegmentation::PropagateComponent(const TriFace& face, 
										  std::map<TriFace, int>&  faceIndices,
										  const std::set<Edge3D>&  featureEdges, 
										  int cur_ID) 
{
	std::vector<TriFace>&   facets = m_pMesh->m_faces;
	std::set<Edge3D>&       edges = m_pMesh->m_edges;

	std::set<Edge3D>::const_iterator pos;
	int index0, index1, index2;
	bool found = false;
	Edge3D edge;

	std::stack<TriFace> stack;
	stack.push(face);

	while( !stack.empty() ) {
		TriFace top = stack.top() ;

		stack.pop() ;

		int index = faceIndices[top];
		if( m_faceID[ index ]  == -1 ) {
			m_faceID[ index ] = cur_ID;

			index0 = top.index0;
			index1 = top.index1;
			index2 = top.index2;

			edge = Edge3D(index0, index1);
			found = (featureEdges.find(edge) != featureEdges.end());
			if ( !found )
			{
				pos = edges.find(edge);
				edge = *pos;
				if (!edge.isBorder)
				{
					if (m_faceID[edge.neighbor0] == -1)
						stack.push(facets[edge.neighbor0]);
					if (m_faceID[edge.neighbor1] == -1)
						stack.push(facets[edge.neighbor1]);
				}
			}
			edge = Edge3D(index1, index2);
			found = (featureEdges.find(edge) != featureEdges.end());
			if ( !found )
			{
				pos = edges.find(edge);
				edge = *pos;
				if (!edge.isBorder)
				{
					if (m_faceID[edge.neighbor0] == -1)
						stack.push(facets[edge.neighbor0]);
					if (m_faceID[edge.neighbor1] == -1)
						stack.push(facets[edge.neighbor1]);
				}
			}
			edge = Edge3D(index2, index0);
			found = (featureEdges.find(edge) != featureEdges.end());
			if ( !found )
			{
				pos = edges.find(edge);
				edge = *pos;
				if (!edge.isBorder)
				{
					if (m_faceID[edge.neighbor0] == -1)
						stack.push(facets[edge.neighbor0]);
					if (m_faceID[edge.neighbor1] == -1)
						stack.push(facets[edge.neighbor1]);
				}
			}
		}
	}
}   


// 淨ڻ
double MeshSegmentation::FaceNormalProduct(const Edge3D& edge) const 
{
	assert(!edge.isBorder);

	int tri0 = edge.neighbor0;
	int tri1 = edge.neighbor1;

	return (m_pMesh->m_faceNormals[tri0].DotProduct(m_pMesh->m_faceNormals[tri1]));
}

// 淨н
double MeshSegmentation::FaceNormalAngle(const Edge3D& edge) const 
{
	assert(!edge.isBorder);

	int tri0 = edge.neighbor0;
	int tri1 = edge.neighbor1;

	double product = m_pMesh->m_faceNormals[tri0].DotProduct(m_pMesh->m_faceNormals[tri1]);
	return acos(product) / DOUBLE_PI * 180;
}
