#include "StdAfx.h"
#include "PolygonDelaunayGraph.h"
#include "pdg_insert.h"

#include <cgal/squared_distance_2.h>



PolygonDelaunayGraph::PolygonDelaunayGraph(void)
: m_bShowGraph(true)
, m_bShowNarrowAreas(true)
, m_dNarrowThreshold(0)
, m_nDisplayListNarrowAreas(0)
{}


void PolygonDelaunayGraph::SetPolygonData(const std::list<std::list<Point_2>>& polygons)
{
	if ( polygons.size() > 0 )
	{
		Clear();

		// convert to CGAL polygons
		std::list< std::list<Point_2> >::const_iterator polygon_cur = polygons.begin();
		std::list< std::list<Point_2> >::const_iterator polygon_end = polygons.end();

		for (; polygon_cur!=polygon_end; ++polygon_cur)
		{
			Polygon_2 polygon(polygon_cur->begin(), polygon_cur->end());
			m_polygons.push_back(polygon);
		}

		ComputeGeometryParameters(polygons);
	}
}


void PolygonDelaunayGraph::Create()
{
	size_t size = m_polygons.size();
	if ( size > 0 )
	{
		for ( size_t i=0; i<size; ++i )
			insert_polygon(m_sdg, m_polygons[i]);
		
		// validate the segment Delaunay graph
		if (m_sdg.is_valid(true, 1)) {
			BuildDisplayListObject();
			FindAllPointsAndTheirSites();
		}
		else
			AfxMessageBox(_T("Construct polygon Delaunay graph failed."));
	}
}


void PolygonDelaunayGraph::Clear()
{
	m_sdg.clear();
	m_polygons.clear();

	m_allPointsAndNearestSites.clear();
	m_narrowAreaPointsAndSites.clear();

	glDeleteLists(m_nDisplayListObject, 1);
	glDeleteLists(m_nDisplayListNarrowAreas, 1);
	m_nDisplayListObject = 0;
	m_nDisplayListNarrowAreas = 0;
}


void PolygonDelaunayGraph::SetShowGraph(bool b)
{
	m_bShowGraph = b;
}


void PolygonDelaunayGraph::SetShowNarrowAreas(bool b)
{
	m_bShowNarrowAreas = b;
}


void PolygonDelaunayGraph::Draw() 
{
	if (m_bShowGraph)
		glCallList(m_nDisplayListObject);

	if (m_bShowNarrowAreas)
		glCallList(m_nDisplayListNarrowAreas);
}

void PolygonDelaunayGraph::BuildDisplayListNarrowAreas()
{
	m_nDisplayListNarrowAreas = glGenLists(1);
	glNewList(m_nDisplayListNarrowAreas, GL_COMPILE);
		::glPushMatrix();
		::glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT |  GL_POINT_BIT  | GL_LIGHTING_BIT);
		::glDisable(GL_LIGHTING);

		::glScalef(m_fFitScale, m_fFitScale, m_fFitScale);
		::glTranslated(-m_center.x(), -m_center.y(), 0);

		::glPointSize(10.0f);
		::glColor3ub(255, 0, 0);
		// Voronoi Diagramϵĵ
		::glBegin(GL_POINTS);
			for (size_t i=0; i<m_narrowAreaPointsAndSites.size(); ++i)
			{
				Point_2 p = m_narrowAreaPointsAndSites[i].first;
				glVertex2d(p.x(), p.y());
			}
		::glEnd();

		::glLineWidth(7.0f);
		::glColor3ub(0, 255, 0);		
		::glBegin(GL_LINES);
		for (size_t i=0; i<m_narrowAreaPointsAndSites.size(); ++i)
		{
			PointAndSite& ps = m_narrowAreaPointsAndSites[i];
			SDG_2::Site_2 site = ps.second;

			if (site.is_segment())  // Σеsite߶
			{
				Segment_2 s = site.segment();
				Point_2  p0 = s.source();
				Point_2  p1 = s.target();

				glVertex2d( p0.x(), p0.y() );	
				glVertex2d( p1.x(), p1.y() );	
			}
		}
		::glEnd();
		
		::glPopAttrib();
		::glPopMatrix();
	::glEndList();
}

void PolygonDelaunayGraph::BuildDisplayListObject()
{
	m_nDisplayListObject = glGenLists(1);
	glNewList(m_nDisplayListObject, GL_COMPILE);
		::glPushMatrix();
		::glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_LINE_BIT | GL_LIGHTING_BIT);

		::glScalef(m_fFitScale, m_fFitScale, m_fFitScale);
		::glTranslated(-m_center.x(), -m_center.y(), 0);

		::glDisable(GL_LIGHTING);
		::glLineWidth(2.0f);
		::glColor3ub(0, 255, 255);

		SDG_2::Finite_edges_iterator eit = m_sdg.finite_edges_begin();
		for (; eit != m_sdg.finite_edges_end(); ++eit) 
		{
			int id0 = eit->first->vertex( m_sdg.cw(eit->second) )->info();
			int id1 = eit->first->vertex( m_sdg.ccw(eit->second) )->info();
			if ( id0 != id1 )
				DrawEdge(eit);	
		}

		::glPopAttrib();
		::glPopMatrix();
	glEndList();
}

void PolygonDelaunayGraph::FindAllPointsAndTheirSites()
{
	// Voronoi diagram е㼰Site
	SDG_2::Finite_edges_iterator eit = m_sdg.finite_edges_begin();
	for (; eit != m_sdg.finite_edges_end(); ++eit) 
	{
		int id0 = eit->first->vertex( m_sdg.cw(eit->second) )->info();
		int id1 = eit->first->vertex( m_sdg.ccw(eit->second) )->info();
		
		SDG_2::Edge e = *eit;
		if ( (id0 != id1) && (!m_sdg.is_infinite(e)) )
		{
			Gt::Line_2						l;
			Gt::Segment_2					s;
			Gt::Ray_2						r;
			CGAL::Parabola_segment_2<Gt>    ps;

			CGAL::Object o = m_sdg.primal(e);

			if (CGAL::assign(l, o))	{
				//line;
			}

			if (CGAL::assign(s, o)) {
				RecordPoint(s.source());
				RecordPoint(s.target());
			}

			if (CGAL::assign(r, o))
				RecordPoint(r.source());

			if (CGAL::assign(ps, o)) 
			{
				std::vector<Point_2> points;
				ps.generate_points(points);
				for (unsigned int i = 0; i < points.size(); i++)		
					RecordPoint(points[i]);
			}
		}
	}
}


void PolygonDelaunayGraph::RecordPoint(Point_2 p)
{
	SDG_2::Vertex_handle vh = m_sdg.nearest_neighbor(p);
	SDG_2::Site_2 site = vh->site();
	m_allPointsAndNearestSites[p] = site;
}


void PolygonDelaunayGraph::DetectNarrowAreas(double narrowThreshold)
{
	if (m_allPointsAndNearestSites.empty())	{
		AfxMessageBox(_T("ȼPolygon Delaunay Graph!"));
		return;
	}

	m_narrowAreaPointsAndSites.clear();
	m_dNarrowThreshold = narrowThreshold / 2.0;

	std::map<Point_2, SDG_2::Site_2>::const_iterator cur = m_allPointsAndNearestSites.begin();
	std::map<Point_2, SDG_2::Site_2>::const_iterator end = m_allPointsAndNearestSites.end();
	for (; cur!=end; ++cur)
	{
		Point_2 p0 = cur->first;

		SDG_2::Site_2 site = cur->second;
		
		double sqrdDistance = FLOAT_MAX;
		if (site.is_point()) {
			Point_2 p1 = site.point();
			sqrdDistance = CGAL::squared_distance(p0, p1);
		}
		else if (site.is_segment())	{
			Segment_2 s = site.segment();
			sqrdDistance = CGAL::squared_distance(p0, s);
		}

		if (sqrdDistance <= m_dNarrowThreshold * m_dNarrowThreshold)
			m_narrowAreaPointsAndSites.push_back(*cur);
	}

	BuildDisplayListNarrowAreas();
}


void PolygonDelaunayGraph::DrawEdge(const SDG_2::Finite_edges_iterator& eit) 
{
	SDG_2::Edge e = *eit;

	if (m_sdg.is_infinite(e))
		return;

	Gt::Line_2						l;
	Gt::Segment_2					s;
	Gt::Ray_2						r;
	CGAL::Parabola_segment_2<Gt>    ps;

	CGAL::Object o = m_sdg.primal(e);

	if (CGAL::assign(l, o))
	{
		//drawLine(l);
	}

	if (CGAL::assign(s, o)) 
	{
		Point_2 p0 = s.source();  
		Point_2 p1 = s.target(); 
		glColor3ub(255, 255, 0);
		glBegin(GL_LINES);
			glVertex2d( p0.x(), p0.y() );	
			glVertex2d( p1.x(), p1.y() );	
		glEnd();
	}

	if (CGAL::assign(r, o))  
	{
		Point_2 p0 = r.source(); 
		Point_2 p1 = r.point(1);
		glColor3ub(255, 255, 0);
		glBegin(GL_LINES);
		glVertex2d( p0.x(), p0.y() );	
		glVertex2d( p1.x(), p1.y() );	
		glEnd();
	}

	if (CGAL::assign(ps, o)) 
	{
		std::vector<Point_2> points;
		ps.generate_points(points);

		glColor3ub(255, 255, 0);
		glBegin(GL_LINE_STRIP);
		for (unsigned int i = 0; i < points.size(); i++)
		{
			Point_2 p = points[i];	 		
			glVertex2d( p.x(), p.y() );
		}
		glEnd();	
	}
}

