﻿/*
* Copyright (c) 2007, Shenyang Institute of Automation, Chinese Academy
* of Sciences. 南亮亮, all rights reserved.
* 
* Filename     : TriMesh.cpp
* Project	   : xMesh library
* Abstract     : Implementation of the TriMesh class
* Version      : 1.0
* Author			: 南亮亮, nanlliang@sia.cn
* Original Author   : 南亮亮, nanlliang@sia.cn
* Accomplished date : Dec. 29, 2007
*/

//-----------------------------------------------------------------

#include "StdAfx.h"

#include <io.h>
#include <fcntl.h>
#include <map>
#include <map>
#include <algorithm>
using namespace std;

#include "trimesh.h"
#include "Constant.h"


TriMesh::TriMesh()
: m_faceNum(0)
, m_vertexNum(0)
, m_fitScale(1.0)
, m_bShowModel(true)
, m_bShowFeatureEdges(false)
{
}

TriMesh::~TriMesh(void)
{
	Clear();
}


void TriMesh::Clear()
{
	m_vertices.clear();
	m_faces.clear();
	m_edges.clear();
	m_vertexNormals.clear();
	m_faceNormals.clear();
	m_boundaries.clear();
}

void TriMesh::SetVertexNumber(int vertex_N)
{
	if(m_vertices.size() > 0)
		m_vertices.clear();
	m_vertexNum = vertex_N;
	m_vertices.resize(vertex_N);
}

void TriMesh::SetFaceNumber(int face_N)
{
	if(m_faces.size() > 0)
		m_faces.clear();
	m_faceNum = face_N;
	m_faces.resize(face_N);
}


void TriMesh::ComputeGeoParameters()
{
	m_bboxMin.x = m_bboxMin.y = m_bboxMin.z = FLOAT_MAX;
	m_bboxMax.x = m_bboxMax.y = m_bboxMax.z = -FLOAT_MAX;

	for (int i=0; i<m_vertexNum; i++) {
		m_bboxMin.x = m_vertices[i].x < m_bboxMin.x ? m_vertices[i].x : m_bboxMin.x;
		m_bboxMin.y = m_vertices[i].y < m_bboxMin.y ? m_vertices[i].y : m_bboxMin.y;
		m_bboxMin.z = m_vertices[i].z < m_bboxMin.z ? m_vertices[i].z : m_bboxMin.z;

		m_bboxMax.x = m_vertices[i].x > m_bboxMax.x ? m_vertices[i].x : m_bboxMax.x;
		m_bboxMax.y = m_vertices[i].y > m_bboxMax.y ? m_vertices[i].y : m_bboxMax.y;
		m_bboxMax.z = m_vertices[i].z > m_bboxMax.z ? m_vertices[i].z : m_bboxMax.z;
	}

	m_center.x = ( m_bboxMax.x + m_bboxMin.x ) / 2.0;
	m_center.y = ( m_bboxMax.y + m_bboxMin.y ) / 2.0;
	m_center.z = ( m_bboxMax.z + m_bboxMin.z ) / 2.0;

	// Calculate proper rendering position
	 double len = max( m_bboxMax.x - m_bboxMin.x,
		max(m_bboxMax.y - m_bboxMin.y, m_bboxMax.z - m_bboxMin.z) );
	m_fitScale = 1.5 / len;
}


void TriMesh::ComputeFaceNormal()
{
	if(m_faceNormals.size() > 0)
		m_faceNormals.clear();
	m_faceNormals.resize(m_faceNum);

	int i0 = 0;
	int i1 = 0;
	int i2 = 0;
	for(int i=0; i<m_faceNum; i++) {
		i0 = m_faces[i].index0;
		i1 = m_faces[i].index1;
		i2 = m_faces[i].index2;

		if (i0 < 0 || i1 < 0 || i2 < 0)
			continue;

		Vector3D c = Vector3D(m_vertices[i0], m_vertices[i1]) * Vector3D(m_vertices[i0], m_vertices[i2]);
		c.Normalize();
		m_faceNormals[i] = c;
	}
}

void TriMesh::ComputeVertexNormal() 
{
	if(m_vertexNormals.size() > 0)
		m_vertexNormals.clear();
	m_vertexNormals.resize(m_vertexNum);

	vector<double>  sumWeight(m_vertexNum, 1);

	for(int i=0; i<m_faceNum; i++) {
		int i0 = m_faces[i].index0;
		int i1 = m_faces[i].index1;
		int i2 = m_faces[i].index2;

		if (i0 < 0 || i1 < 0 || i2 < 0)
			continue;

		double area = FaceArea(m_vertices[i0], m_vertices[i1], m_vertices[i2]);
		Vector3D normal = Vector3D(m_vertices[i0], m_vertices[i1]) * Vector3D(m_vertices[i0], m_vertices[i2]);
		
		//////////////////////////////////////////////////////////////////////////
		// 相邻面的面发向量面积加权
		//m_vertexNormals[i0] += normal * area;
		//sumWeight[i0] += area;

		//m_vertexNormals[i1] += normal * area;
		//sumWeight[i1] += area;

		//m_vertexNormals[i2] += normal * area;
		//sumWeight[i2] += area;

		///////////////////////////////////////////////////////////////////////////
		// 王华兵: 面法向量的(角度/面积)加权平均
		Vector3D v0 = Vector3D(m_vertices[i0], m_vertices[i1]);
		Vector3D v1 = Vector3D(m_vertices[i0], m_vertices[i2]);			
		double alpha = acos( v0.DotProduct(v1) / (v0.Length() * v1.Length()) );
		double weight = alpha / area;
		m_vertexNormals[i0] += normal * weight;
		sumWeight[i0] += weight;

		v0 = Vector3D(m_vertices[i1], m_vertices[i0]);
		v1 = Vector3D(m_vertices[i1], m_vertices[i2]);			
		alpha = acos( v0.DotProduct(v1) / (v0.Length() * v1.Length()) );
		weight = alpha / area;
		m_vertexNormals[i1] += normal * weight;
		sumWeight[i1] += weight;

		v0 = Vector3D(m_vertices[i2], m_vertices[i1]);
		v1 = Vector3D(m_vertices[i2], m_vertices[i0]);			
		alpha = acos( v0.DotProduct(v1) / (v0.Length() * v1.Length()) );
		weight = alpha / area;
		m_vertexNormals[i2] += normal * weight;
		sumWeight[i2] += weight;
	}

	double length = 0;
	for(int i=0; i<m_vertexNum; i++){
		m_vertexNormals[i] /= sumWeight[i];
		m_vertexNormals[i].Normalize();
	}
}

void TriMesh::ExtractEdges()
{
	if (!m_edges.empty())
		m_edges.clear();

	int index0 = 0;
	int index1 = 0;
	int index2 = 0;

	set<Edge3D>::iterator pos;
	for (int i=0; i<m_faceNum; i++)	{
		index0 = m_faces[i].index0;
		index1 = m_faces[i].index1;
		index2 = m_faces[i].index2;

		Edge3D e0(index1, index2);
		pos = m_edges.find(e0);
		if ( pos != m_edges.end() ) {
			pos->neighbor1 = i;
			pos->isBorder = false;
		}
		else {
			e0.neighbor0 = i;
			m_edges.insert(e0);
		}

		Edge3D e1(index0, index2);
		pos = m_edges.find(e1);
		if ( pos != m_edges.end() ) {
			pos->neighbor1 = i;
			pos->isBorder = false;
		}
		else {
			e1.neighbor0 = i;
			m_edges.insert(e1);
		}

		Edge3D e2(index0, index1);
		pos = m_edges.find(e2);
		if ( pos != m_edges.end() ) {
			pos->neighbor1 = i;
			pos->isBorder = false;
		}
		else {
			e2.neighbor0 = i;
			m_edges.insert(e2);
		}
	}

	set<Edge3D>::const_iterator cur = m_edges.begin();
	set<Edge3D>::const_iterator end = m_edges.end();
	for (; cur!=end; cur++)	{
		if ((!cur->isBorder)  &&  (cur->neighbor0 < 0 || cur->neighbor1 < 0 ))
		{
			AfxMessageBox(_T("Error!"));
		}

		if ( (cur->isBorder) && (cur->neighbor0 < 0 && cur->neighbor1 > 0) ) 
		{
			AfxMessageBox(_T("Error!"));
		}
	}
}

void TriMesh::Normalize()
{
	double len = max( m_bboxMax.x - m_bboxMin.x,
		max(m_bboxMax.y - m_bboxMin.y, m_bboxMax.z - m_bboxMin.z) );

	for(int i=0; i<m_vertexNum; i++)
	{
		m_vertices[i] -= m_center;
		m_vertices[i] /= len;
	}

	ComputeGeoParameters();
}


void TriMesh::ReverseInOut()
{
	for (int i=0; i<m_faceNum; ++i)
		swap(m_faces[i].index0, m_faces[i].index1);		
}

double TriMesh::FaceArea(Point3D p0, Point3D p1, Point3D p2) const {
	Vector3D v = Vector3D(p0, p1) * Vector3D(p0, p2);
	return 0.5 * v.Length();
}

double TriMesh::FaceArea(int faceIndex) const {
	TriFace face = m_faces[faceIndex];		
	int i0 = face.index0;
	int i1 = face.index1;
	int i2 = face.index2;

	if (i0 < 0 || i1 < 0 || i2 < 0)
		return 0;

	Point3D p0 = m_vertices[i0];
	Point3D p1 = m_vertices[i1];
	Point3D p2 = m_vertices[i2];

	return FaceArea(p0, p1, p2);
}

Point3D TriMesh::FaceCenter(int faceIndex) const
{
	TriFace face = m_faces[faceIndex];

	int index0 = face.index0;
	int index1 = face.index1;
	int index2 = face.index2;

	double x = (m_vertices[index0].x + m_vertices[index1].x + m_vertices[index2].x) / 3.0;
	double y = (m_vertices[index0].y + m_vertices[index1].y + m_vertices[index2].y) / 3.0;
	double z = (m_vertices[index0].z + m_vertices[index1].z + m_vertices[index2].z) / 3.0;

	return Point3D(x, y, z);
}

Vector3D TriMesh::FaceNormal(int faceIndex) const
{
	TriFace face = m_faces[faceIndex];

	int index0 = face.index0;
	int index1 = face.index1;
	int index2 = face.index2;

	Vector3D normal = Vector3D(m_vertices[index0], m_vertices[index1]) * 
		Vector3D(m_vertices[index0], m_vertices[index2]);

	normal.Normalize();

	return normal;
}


bool TriMesh::OpenSTL(CString filename)
{
	//pFile = fopen(file_name, "r");
	FILE*   pFile;
	errno_t err = _wfopen_s(&pFile, filename, _T("rt"));
	if( err )
	{
		AfxMessageBox(_T("The file was not opened!"));
		return false;
	}

	// first read max(400, fileSize) bytes to determine whether it is a binary or ascii file
	char strRead[400];
	fread(strRead, sizeof(char), 400, pFile);
	rewind(pFile);

	// if "strRead" contains "normal", then it is a ascii file
	char* isAscii = strstr(strRead, "normal");

	if (isAscii)
		OpenSTLAscii(pFile);
	else
	{
		_setmode( _fileno(pFile), _O_BINARY ); // change to binary mode
		OpenSTLBinary(pFile);
	}

	fclose(pFile);

	ComputeGeoParameters();

	return true;
}

bool TriMesh::SaveSTL(CString filename)
{
	FILE* pFile;
	errno_t err = _wfopen_s(&pFile, filename, _T("wt"));
	if( err )
	{
		AfxMessageBox(_T("The file was not opened!"));
		return false;
	}

	CString fileExt = filename.Right(filename.GetLength() - filename.ReverseFind('.') - 1);
	fileExt.MakeLower();

	if (fileExt == "stl" || fileExt == "astl" || fileExt == "stla")
		SaveSTLAscii(pFile);
	else if (fileExt == "bstl" || fileExt == "stlb")
	{
		_setmode( _fileno(pFile), _O_BINARY ); // change to binary mode
		SaveSTLBinary(pFile);
	}

	fclose(pFile);

	return true;
}

void TriMesh::OpenSTLAscii(FILE* pFile)
{
	Triangle tri;
	vector<Triangle> stlModel;

	long fileLenth = _filelength(_fileno(pFile));

	char str[80];
	while(fscanf_s(pFile, "%s", str, 80) != EOF)
	{
		if(strncmp(str, "normal", 6) == 0)
		{
			// face normal
			fscanf_s(pFile, "%lf %lf %lf", &(tri.normal.x), &(tri.normal.y), &(tri.normal.z));

			// skip 
			fscanf_s(pFile, "%*s %*s");

			// three vertices
			for (unsigned int i = 0; i<3; i++)
				fscanf_s(pFile, "%*s %lf %lf %lf", &(tri.vertices[i].x), &(tri.vertices[i].y), &(tri.vertices[i].z));

			stlModel.push_back(tri);
		}  		
	}

	STL2TriMesh(stlModel);
}


void TriMesh::OpenSTLBinary(FILE* pFile)
{
	if ( fseek(pFile, 80L, SEEK_SET) != 0 )
	{
		AfxMessageBox(_T("Error: file is less than 80 bytes!"));
		return;
	}

	int triNum = 0;
	fread(&triNum, 4, 1, pFile); // read face number

	long beginPos = ftell(pFile);
	fseek(pFile, 0L, SEEK_END);
	long endPos = ftell(pFile);

	long sNeeded = triNum * 50;       // 50 bytes per face
	long sAvail  = endPos - beginPos;

	if (sNeeded != sAvail)
	{
		CString msg;
		msg.Format(_T("Error:\n STL file's triangle count does not match file size.\n Bytes needed: %ld, available data: %ld"), 
			sNeeded, sAvail);
		AfxMessageBox(msg);
		return;
	}

	rewind(pFile);
	fseek(pFile, 84L, SEEK_SET); // 84 = 80 + 4

	Triangle tri;
	vector<Triangle> stlModel;

	float ratio = 100.0 / triNum; 

	float x, y, z;
	for (int i=0; i<triNum; i++)
	{
		// face normal
		fread( &x, 4, 1, pFile );	tri.normal.x = x;
		fread( &y, 4, 1, pFile );	tri.normal.y = y;
		fread( &z, 4, 1, pFile );	tri.normal.z = z;

		// three vertices
		for (int j=0; j<3; j++)
		{
			fread( &x, 4, 1, pFile );	tri.vertices[j].x = x;
			fread( &y, 4, 1, pFile );	tri.vertices[j].y = y;
			fread( &z, 4, 1, pFile );	tri.vertices[j].z = z;
		}

		stlModel.push_back(tri);

		fseek(pFile, 2L, SEEK_CUR);
	}

	STL2TriMesh(stlModel);
}


void TriMesh::SaveSTLAscii(FILE* pFile)
{
	fprintf_s(pFile, "Solid MeshShop Part. Generated by Liangliang Nan\n");
	int index0, index1, index2;
	for (int i=0; i<m_faceNum; i++)
	{
		index0 = m_faces[i].index0;
		index1 = m_faces[i].index1;
		index2 = m_faces[i].index2;

		fprintf_s(pFile, "   facet normal ");
		fprintf_s(pFile, "%f %f %f\n", m_faceNormals[i].x, m_faceNormals[i].y, m_faceNormals[i].z);
		fprintf_s(pFile, "      outer loop\n");
		fprintf_s(pFile, "         vertex %f %f %f\n", m_vertices[index0].x, m_vertices[index0].y, m_vertices[index0].z);
		fprintf_s(pFile, "         vertex %f %f %f\n", m_vertices[index1].x, m_vertices[index1].y, m_vertices[index1].z);
		fprintf_s(pFile, "         vertex %f %f %f\n", m_vertices[index2].x, m_vertices[index2].y, m_vertices[index2].z);
		fprintf_s(pFile, "      endloop\n   endfacet\n");
	}
	fprintf_s(pFile, "%s", "end solid");
}

void TriMesh::SaveSTLBinary(FILE* pFile)
{
	fprintf_s(pFile, "Solid MeshShop Part. Generated by Liangliang Nan\n");

	rewind(pFile);
	fseek(pFile, 80L, SEEK_SET);

	int triNum = m_faceNum;
	fwrite(&triNum, 4, 1, pFile);

	if (m_faceNormals.empty())
		ComputeFaceNormal();

	char ignor[2];
	float x, y, z;
	TriFace  face;
	for (int i=0; i<triNum; i++)
	{
		x = (float)m_faceNormals[i].x;   fwrite(&x, 4, 1, pFile );
		y = (float)m_faceNormals[i].y;   fwrite(&y, 4, 1, pFile );
		z = (float)m_faceNormals[i].z;   fwrite(&z, 4, 1, pFile );

		face = m_faces[i];
		for(int j=0; j<3; j++)
		{
			x = (float)m_vertices[ face[j] ].x;	fwrite(&x, 4, 1, pFile );
			y = (float)m_vertices[ face[j] ].y;   fwrite(&y, 4, 1, pFile );
			z = (float)m_vertices[ face[j] ].z;   fwrite(&z, 4, 1, pFile );
		}

		fwrite(&ignor, 2, 1, pFile);
	}
}


void TriMesh::STL2TriMesh(vector<Triangle> &triangles)
{
	typedef pair<Point3D, int> PtAndIndex;

	map< Point3D, int, less<Point3D> > temp;
	PtAndIndex ptAndIndex;
	TriFace face;
	Point3D point;

	int triNum = (int)triangles.size();

	for (int i=0; i<triNum; i++)
	{
		// face vertex 0
		point = triangles[i].vertices[0];
		if (temp.find(point) != temp.end())
			face.index0 = temp[point];
		else
		{
			ptAndIndex.first = point;
			ptAndIndex.second = (int)temp.size();
			face.index0 = ptAndIndex.second;
			temp.insert(ptAndIndex);
		}

		// face vertex 1
		point = triangles[i].vertices[1];	
		if (temp.find(point) != temp.end())
			face.index1 = temp[point];
		else
		{
			ptAndIndex.first = point;
			ptAndIndex.second = (int)temp.size();
			face.index1 = ptAndIndex.second;
			temp.insert(ptAndIndex);
		}

		// face vertex 2
		point = triangles[i].vertices[2];
		if (temp.find(point) != temp.end())
			face.index2 = temp[point];
		else
		{
			ptAndIndex.first = point;
			ptAndIndex.second = (int)temp.size();
			face.index2 = ptAndIndex.second;
			temp.insert(ptAndIndex);
		}

		bool faceValid = true;
		for(int j=0; j<3; j++)
		{
			if(face[j] == face[(j+1)%3])
				faceValid = false;
		}
		if (faceValid)
		{
			m_faces.push_back(face);
			m_faceNormals.push_back(triangles[i].normal);
		}
	}  

	m_faceNum = (int)m_faces.size();

	// sort by vertex indices;
	typedef pair<int, Point3D> IndexAndPt;
	map< int, Point3D, less<int> > indicesAndValue;

	IndexAndPt indexAndPt;

	map< Point3D, int, less<Point3D> > :: iterator curTemp = temp.begin();
	map< Point3D, int, less<Point3D> > :: iterator endTemp = temp.end();
	for (; curTemp!=endTemp; curTemp++)
	{
		indexAndPt.first = curTemp->second;
		indexAndPt.second = curTemp->first;	

		indicesAndValue.insert(indexAndPt);
	}

	int vNum = (int)indicesAndValue.size();

	SetVertexNumber(vNum);

	// copy all vertices
	for (int i=0; i<vNum; i++)
		SetVertex(i, indicesAndValue[i]);
}


void TriMesh::DrawModel()
{
	if ((!m_bShowModel) || m_faces.empty())
		return;

	::glPushMatrix();
	::glPushAttrib(GL_CURRENT_BIT | GL_POINT_BIT | GL_LINE_BIT | GL_LIGHTING_BIT);
	
	::glScaled(m_fitScale, m_fitScale, m_fitScale);
	::glTranslated(-m_center.x, -m_center.y, -m_center.z);

	if (m_faceNormals.empty())
		ComputeFaceNormal();

	int i0 = 0; int i1 = 0; int i2 = 0;

	::glPointSize(2.0f);
	::glLineWidth(1.0f);
	::glShadeModel(GL_FLAT);
	::glBegin(GL_TRIANGLES);
		for(int i=0; i<m_faceNum; i++){
			i0 = m_faces[i].index0;
			i1 = m_faces[i].index1;
			i2 = m_faces[i].index2;
			if (i0 < 0 || i1 < 0 || i2 < 0)
				continue;
			::glNormal3d(m_faceNormals[i].x, m_faceNormals[i].y, m_faceNormals[i].z);
			::glVertex3d(m_vertices[i0].x, m_vertices[i0].y, m_vertices[i0].z);
			::glVertex3d(m_vertices[i1].x, m_vertices[i1].y, m_vertices[i1].z);
			::glVertex3d(m_vertices[i2].x, m_vertices[i2].y, m_vertices[i2].z);
		}
	::glEnd();

	::glPopAttrib();
	::glPopMatrix();
}


void TriMesh::DrawFeatureEdges()
{
	if ((!m_bShowFeatureEdges) || m_faces.empty() || m_featureEdges.empty())
		return;

	::glPushMatrix();
	::glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_LIGHTING_BIT | GL_LINE_BIT);
	::glScaled(m_fitScale, m_fitScale, m_fitScale);
	::glTranslated(-m_center.x, -m_center.y, -m_center.z);

	::glDisable(GL_LIGHTING);
	::glShadeModel(GL_SMOOTH);
	::glLineWidth(2.0f);
	::glColor3ub(255, 0, 0);

	int size = (int)m_featureEdges.size();
	::glBegin(GL_LINES);
	for (int i=0; i<size; i++)
	{
		int index0 = m_featureEdges[i].index0;
		int index1 = m_featureEdges[i].index1;
		::glVertex3f(m_vertices[index0].x, m_vertices[index0].y, m_vertices[index0].z);
		::glVertex3f(m_vertices[index1].x, m_vertices[index1].y, m_vertices[index1].z);
	}
	::glEnd();

	::glPopAttrib();
	::glPopMatrix();
}

void TriMesh::SetShowModel(bool b)
{
	m_bShowModel = b;
}

void TriMesh::SetShowFeatureEdges(bool b)
{
	m_bShowFeatureEdges = b;
}