
/*
* Copyright (c) 2007, Shenyang Institute of Automation, Chinese Academy
* of Sciences. , all rights reserved.
* 
* Filename     : TriMesh.h
* Project	   : xMesh library
* Abstract     : Declaration of TriMesh class
* Version      : 1.0
* Author			: , nanlliang@sia.cn
* Original Author   : , nanlliang@sia.cn
* Accomplished date : Dec. 29, 2007
*/

//-----------------------------------------------------------------

#ifndef _TRIMESH_H_
#define _TRIMESH_H_


#include <cmath>
#include <set>
#include <vector>


#include <gl/glut.h>


#include "Point2D.h"
#include "Point3D.h"
#include "Vector3D.h"
#include "Triangle.h"


// vertex types
enum VertexType {
	VT_INNER,
	VT_BOUNDARY,
	VT_NON_MANIFOLD
};

//////////////////////////////////////////////////////////////////////////
// һ, Edge3D(1, 2)  Edge3D(2, 1) ͬһߣ
// ص㣺 index0  С index1
class Edge3D
{
public:
	int index0;
	int index1;
	
	bool isBorder;    // θ
	int  neighbor0;   // 
	int  neighbor1;

public:
	Edge3D() : index0(0), index1(0), isBorder(true), neighbor0(-1), neighbor1(-1)
	{};

	Edge3D(const Edge3D& e) {
		index0 = e.index0;
		index1 = e.index1; 
		isBorder = e.isBorder;
		neighbor0 = e.neighbor0;
		neighbor1 = e.neighbor1;
	}

	Edge3D(int i0, int i1) : isBorder(true), neighbor0(-1), neighbor1(-1) {
		//assert((i0 != i1) && (i0 > 0) && (i1 > 0));
		index0 = min(i0, i1);
		index1 = max(i0, i1);
	}

	void operator=(const Edge3D& edge) {
		index0 = edge.index0;
		index1 = edge.index1;	
		isBorder = edge.isBorder;
		neighbor0 = edge.neighbor0;
		neighbor1 = edge.neighbor1;
	}

	bool operator<(const Edge3D& edge) const {
		return index0 < edge.index0 || ((index0 == edge.index0) && (index1 < edge.index1));
	}
	bool operator==(const Edge3D& edge) const {
		return (index0 == edge.index0) && (index1 == edge.index1);
	}
	bool operator!=(const Edge3D& edge) const {
		return (index0 != edge.index0) || (index1 != edge.index1);
	}

	~Edge3D() {};
};


// vertex indices of a triangle
class TriFace 
{
public:
	TriFace() : index0(-1), index1(-1), index2(-1) {}
	TriFace(int i0, int i1, int i2) : index0(i0), index1(i1), index2(i2) {
		//assert((i0 > 0) && (i1 > 0) && (i2 > 0));
	}
	~TriFace() {}
	void operator =(const TriFace &face) {index0 = face.index0; index1 = face.index1; index2 = face.index2;}
	
	bool operator<(const TriFace& face) const {
		return (  index0 < face.index0  || 
				((index0 == face.index0) && (index1 < face.index1)) ||
				((index0 == face.index0) && (index1 == face.index1) && (index2 < face.index2)) );
	}

	int operator [] (const int i) const
	{
		switch (i)
		{
		case 0:
			return index0;
		case 1:
			return index1;
		case 2:
			return index2;
		default:
			break;
		}
		return index0;
	}
	int& operator [] (const int i) 
	{
		switch (i)
		{
		case 0:
			return index0;
		case 1:
			return index1;
		case 2:
			return index2;
		default:
			break;
		}
		return index0;
	}
public:
	int index0;
	int index1;
	int index2;
};



class TriMesh
{
public:
	TriMesh(void);
	~TriMesh(void);

public:
	//get & set data
	inline void SetVertex(int index, Point3D point) { m_vertices[index] = point; }
	inline void SetFace(int index, TriFace face) { m_faces[index] = face; }

	void SetFaceNumber(int face_N);
	void SetVertexNumber(int vertex_N);

	void ComputeGeoParameters();  // bounding box, center, normalize scale

	void Normalize();
	void ReverseInOut();

	void ComputeFaceNormal();     // compute normal for every triangular face
	void ComputeVertexNormal();   // compute normal for every vertex (area weighted)

	void ExtractEdges();          // get all edges;

	void GenerateFaceLink();
	void GenerateVertexLink();	

	void GetVertexOneRing(int vertex_index, int *&nei, int &nei_N);  // get one ring of vertex[index];
	bool GetConsistentOneRing(int index, int *&nei_v, int *&nei_f, int &nei_N);

	void Clear();      // called by destructor

	bool OpenSTL(CString filename);			   // STL file
	bool SaveSTL(CString filename);

	void DrawModel();
	void DrawFeatureEdges();

	void SetShowModel(bool b);
	void SetShowFeatureEdges(bool b);

public:
	inline double FaceArea(Point3D p0, Point3D p1, Point3D p2) const;
	inline double FaceArea(int faceIndex) const;
	inline Point3D FaceCenter(int faceIndex) const;
	inline Vector3D FaceNormal(int faceIndex) const;

public:	
	std::vector<Point3D>  m_vertices;       // vertex coordinates
	std::vector<Vector3D> m_vertexNormals;  // vertex normal
	int					  m_vertexNum;      // vertex number
	
	std::vector<TriFace>  m_faces;          // vertex indices of each face
	std::vector<Vector3D> m_faceNormals;    // face normal
	std::vector<int>      m_segmentationID; // ָÿID
	int m_faceNum;                     // face number

	// edges
	std::set<Edge3D>    m_edges;
	std::vector<Edge3D> m_featureEdges;	

	// all boundaries
	std::vector< std::vector<int> > m_boundaries;

	// Bounding Box of the triangle mesh
	Point3D  m_bboxMin;
	Point3D  m_bboxMax;

	//to show object in proper scale
	double   m_fitScale; 

	// Geometric center
	Point3D  m_center;

private:
	bool m_bShowModel;
	bool m_bShowFeatureEdges;

private:
	void OpenSTLAscii(FILE* pFile);
	void OpenSTLBinary(FILE* pFile);
	void SaveSTLAscii(FILE* pFile);
	void SaveSTLBinary(FILE* pFile);
	void STL2TriMesh(std::vector<Triangle> &triangles);
};


#endif